# from xpcs_viwer.viewer_kernel import Viewer_Kernel
from .viewer import run 
from .viewer_kernel import ViewerKernel
from .xpcs_file import XpcsFile as xf
from .xpcs_file import XpcsFile


__all__ = [run, xf, XpcsFile]
__version__ = '0.1.0'
__author__ = 'Miaoqi Chu'
__credits__ = 'Argonne National Laboratory'



# import marisa_trie
import os
from os.path import commonprefix
from .fileIO.hdf_to_str import get_hdf_info
from .fileIO.hdf_reader import get, put, get_type
from .xpcs_file import XpcsFile as xf
import logging
import numpy as np
from collections import deque
from .helper.listmodel import ListDataModel

logger = logging.getLogger(__name__)
pjoin = os.path.join


# the following functions are copied from:
# https://stackoverflow.com/questions/2892931
def long_substr(data):
    substr = ''
    if len(data) > 1 and len(data[0]) > 0:
        for i in range(len(data[0])):
            for j in range(len(data[0]) - i + 1):
                if j > len(substr) and is_substr(data[0][i:i + j], data):
                    substr = data[0][i:i + j]
    return substr


def is_substr(find, data):
    if len(data) < 1 and len(find) < 1:
        return False
    for i in range(len(data)):
        if find not in data[i]:
            return False
    return True


def get_suffix(file_name):
    _, suffix = os.path.splitext(file_name)
    return suffix


def create_id(in_list):
    ret = []
    for x in in_list:
        idx_1 = x.find('_')
        idx_2 = x.rfind('_', 0, len(x))
        idx_3 = x.rfind('_', 0, idx_2)
        ret.append(x[0:idx_1] + x[idx_3:idx_2])

    return ret


def create_id3(in_list, repeat=1, keep_slice=None):
    out_list = [x[::-1] for x in in_list]
    prefix = commonprefix(out_list)
    out_list = [x.replace(prefix, '') for x in out_list]
    out_list = [x[::-1] for x in out_list]
    return out_list


def create_id2(in_list, repeat=1, keep_slice=None):
    """
    :param in_list: input file name list
    :param repeat: number of repeats to remove common string
    :param keep_slice: the slice in the original string to keep, if not given,
        then use the segment before the first underscore.
    :return: label list with minimal information redundancy
    """
    if len(in_list) < 1:
        return []

    if keep_slice is None:
        idx = in_list[0].find('_')
        keep_slice = slice(0, idx + 1)

    keep_str = in_list[0][keep_slice]
    if keep_str[-1] != '_':
        keep_str = keep_str + '_'

    for n in range(repeat):
        substr = long_substr(in_list)
        in_list = [x.replace(substr, '') for x in in_list]

    in_list = [keep_str + x for x in in_list]
    return in_list


class FileLocator(object):
    def __init__(self,
                 path,
                 max_cache_size=None):
        self.path = path
        self.cwd = None
        self.trie = None
        self.source = ListDataModel()
        self.source_search = ListDataModel()
        self.target = ListDataModel()
        self.id_list = None
        self.build(path)
        self.type = None
        self.cache = {}
        if max_cache_size is None:
            # 2G
            self.max_cache_size = 1024 ** 3 * 2

    def get_type(self, fname):
        return get_type(pjoin(self.cwd, fname))

    # def get(self, fname, fields_raw, **kwargs):
    #     return get(pjoin(self.cwd, fname), fields_raw, **kwargs)

    def get_fn_tuple(self, max_points=128, rows=None):
        # compile the filenames upto max_points to a tuple
        if max_points <= 0:
            max_points = len(self.target)

        ret = []
        if rows is None or len(rows) == 0:
            for n in range(min(max_points, len(self.target))):
                ret.append(self.target[n])
        else:
            for n in rows[0:max_points]:
                ret.append(self.target[n])
        return tuple(ret)

    def get_xf_list(self, max_points=8, rows=None):
        """
        get the cached xpcs_file list;
        :param max_points: maxiaml number of xpcs_files to get
        :param rows: a list of index to select; if None is given, then use
            max_points;
        :return: list of xpcs_file objects;
        """
        if max_points <= 0:
            max_points = len(self.target)

        if rows is None or len(rows) == 0:
            selected = list(range(min(max_points, len(self.target))))
        else:
            # make sure no more than max_points are loaded.
            selected = rows[0:max_points]

        ret = []
        for n in selected:
            fn = self.target[n]
            ret.append(self.cache[fn])

        return ret

    def load(self, file_list=None, max_number=1024, progress_bar=None,
             flag_del=True):

        if file_list in [None, []]:
            file_list = self.target

        total_num = min(max_number, len(file_list))
        existing_keys = list(self.cache.keys())

        for n in range(total_num):
            fn = self.target[n]
            if progress_bar is not None:
                progress_bar.setValue((n + 1) / total_num * 100)

            if fn in existing_keys:
                # already exist
                existing_keys.remove(fn)
            else:
                # read from file and output as a dictionary
                try:
                    self.cache[fn] = xf(fn, self.cwd)
                except Exception as e:
                    logger.info("failed to load file: %s", fn)
                    logger.info("%s", str(e))

        if flag_del:
            for key in existing_keys:
                self.cache.pop(key, None)

        return

    def get_hdf_info(self, fname, fstr=None):
        """
        get the hdf information / hdf structure for fname
        :param fname: input filename
        :param fstr: list of filter string;
        :return: list of strings that contains the hdf information;
        """
        if fname in self.cache:
            return self.cache[fname].get_hdf_info(fstr)
        else:
            return ['None']

    def add_target(self, alist, threshold=64):
        if alist in [[], None]:
            return
        if self.type is None:
            self.type = self.get_type(alist[0])

        single_flag = True

        if len(alist) <= threshold:
            for x in alist:
                if x not in self.target:
                    t = self.get_type(x)
                    if t not in ['Multitau', 'Twotime']:
                        logger.info('Failed to get type for %s', x)
                        continue
                    if self.type is None:
                        self.type = t
                    elif t != self.type:
                        logger.info('Mixed analysis type for %s. Discard', x)
                        single_flag = False
                        continue
                    self.target.append(x)

            self.id_list = create_id(self.target)
        else:
            # if many files are added; then ignore the type check;
            logger.info('type check is disabled. too many files added')
            self.target.extend(alist)
            self.id_list = alist.copy()

        logger.info('length of target = %d' % len(self.target))
        return single_flag

    def clear_target(self):
        self.target.clear()
        self.id_list = None
        self.type = None

    def remove_target(self, rlist):
        if rlist is None or len(self.target) == 0:
            return

        for x in rlist:
            if x in self.target:
                self.target.remove(x)
                self.cache.pop(x, 'None')

        if self.target is None or len(self.target) == 0:
            self.clear_target()
        else:
            self.id_list = create_id(self.target)

    def search(self, val, filter_type='prefix'):
        ans = None
        if filter_type == 'prefix':
            ans = [x for x in self.source if x.startswith(val)]
        elif filter_type == 'substr':
            ans = [x for x in self.source if val in x]
        self.source_search.replace(ans)

        return

    def build(self, path=None, filter_list=('.hdf', '.h5')):
        if path is None:
            path = self.path

        if os.path.isfile(path):
            with open(path, 'r') as f:
                flist = [x[:-1] for x in f]
            self.cwd = os.path.dirname(path)
        elif os.path.isdir(path):
            flist = os.listdir(path)
            self.cwd = path
        else:
            return

        flist = [x for x in flist if get_suffix(x) in filter_list]

        # filter configure files
        flist = [x for x in flist if not x.startswith('.')]
        flist.sort()

        self.source.replace(flist)

        return True


def test1():
    fl = FileLocator(path='./data/files.txt')


if __name__ == "__main__":
    # test1()
    fl = FileLocator(path='./data/files.txt')



import numpy as np
from .file_locator import FileLocator
from .module import saxs2d, saxs1d, intt, stability, g2mod, tauq, twotime
from .module.average_toolbox import AverageToolbox
from shutil import copyfile
from sklearn.cluster import KMeans as sk_kmeans
import h5py
from .helper.listmodel import ListDataModel, TableDataModel
import pyqtgraph as pg

import os
import logging

logger = logging.getLogger(__name__)


class ViewerKernel(FileLocator):
    def __init__(self, path, statusbar=None):
        super().__init__(path)
        self.statusbar = statusbar
        self.meta = None
        self.reset_meta()
        self.avg_tb = AverageToolbox(path)
        self.avg_worker = TableDataModel()
        self.avg_jid = 0
        self.avg_worker_active = {}

    def reset_meta(self):
        self.meta = {
            # twotime
            'twotime_fname': None,
            'twotime_dqmap': None,
            'twotime_ready': False,
            'twotime_ims': [],
            'twotime_text': None,
            # avg
            'avg_file_list': None,
            'avg_intt_minmax': None,
            'avg_g2_avg': None,
            # g2
            'g2_num_points': None,
            'g2_data': None,
            'g2_plot_condition': tuple([None, None, None]),
            'g2_fit_val': {}
        }
        return

    def reset_kernel(self):
        self.clear_target()
        self.reset_meta()

    def show_message(self, msg):
        if msg in [None, [None]]:
            return

        if isinstance(msg, list):
            for t in msg:
                logger.info(t)
            msg = '\n'.join(msg)
        else:
            logger.info(msg)

        if self.statusbar is not None:
            self.statusbar.showMessage(msg, 1500)

    def get_g2_data(self, max_points, rows, **kwargs):
        xf_list = self.get_xf_list(max_points, rows=rows)
        flag, tel, qd, g2, g2_err = g2mod.get_data(xf_list, **kwargs)
        return flag, tel, qd, g2, g2_err

    def get_pg_tree(self, rows):
        if rows in [None, []]:
            rows = [0]
        xfile = self.cache[self.target[rows[0]]]
        return xfile.get_pg_tree()
    
    def get_fitting_tree(self):
        xf_list = self.get_xf_list(8) 
        result = {}
        for x in xf_list:
            result[x.label] = x.get_fitting_info(mode='g2_fitting')
        tree = pg.DataTreeWidget(data=result)
        tree.setWindowTitle('fitting summary')
        tree.resize(1024, 800)
        return tree

    def plot_g2(self,
                handler,
                q_range=None,
                t_range=None,
                y_range=None,
                offset=None,
                show_fit=False,
                max_points=128,
                bounds=None,
                show_label=False,
                num_col=4,
                rows=None,
                fit_flag=None,
                y_auto=False,
                subtract_baseline=True,
                plot_type='multiple'):

        fn_tuple = self.get_fn_tuple(max_points, rows=rows)
        new_condition = (
            (fn_tuple, num_col, show_fit, show_label, subtract_baseline),
            (q_range, t_range, y_range, offset, y_auto),
            (bounds, fit_flag))

        if self.meta['g2_plot_condition'] == new_condition:
            # avoid meaningless rerun
            logger.info('g2 plot parameters unchanged; skip')
            return
        else:
            # cmp = tuple(
            # i != j
            # for i, j in zip(new_condition, self.meta['g2_plot_condition']))
            self.meta['g2_plot_condition'] = new_condition
            # plot_level = 4 * cmp[0] + 2 * cmp[1] + cmp[2]
            # logger.info('plot level = %d', plot_level)

        xf_list = self.get_xf_list(max_points, rows=rows) 
        g2mod.pg_plot(handler, xf_list, num_col, q_range, t_range, y_range,
                      offset=offset, show_label=show_label, show_fit=show_fit,
                      bounds=bounds, plot_type=plot_type, fit_flag=fit_flag,
                      subtract_baseline=subtract_baseline, y_auto=y_auto)

        return

    def plot_tauq_pre(self, hdl=None, max_points=128, rows=None):
        xf_list = self.get_xf_list(max_points, rows=rows)
        short_list = [xf for xf in xf_list if xf.fit_summary is not None]
        tauq.plot_pre(short_list, hdl)

    def plot_tauq(self, hdl=None, bounds=None, rows=[], plot_type=3,
                  fit_flag=None, offset=None, max_points=128, q_range=None):
        
        xf_list = self.get_xf_list(max_points, rows=rows) 
        result = {}
        for x in xf_list:
            if x.fit_summary is None:
                logger.info('g2 fitting is not available for %s', x.fname)
            else:
                x.fit_tauq(q_range, bounds, fit_flag)
                result[x.label] = x.get_fitting_info(mode='tauq_fitting')
        
        tauq.plot(xf_list, hdl=hdl, q_range=q_range, offset=offset,
                  plot_type=plot_type)

        return result

    def plot_saxs_2d(self, *args, **kwargs):
        ans = [self.cache[fn].saxs_2d for fn in self.target]
        # extents = extent = (qy_min, qy_max, qx_min, qx_max)
        extent = self.cache[self.target[0]].get_detector_extent()
        saxs2d.plot(ans, extent=extent, *args, **kwargs)

    def plot_saxs_1d(self, mp_hdl, max_points=128, **kwargs):
        xf_list = self.get_xf_list(max_points)
        saxs1d.plot(xf_list, mp_hdl, max_points=max_points, 
                    legend=self.id_list, **kwargs)

    def setup_twotime(self, file_index=0, group='xpcs'):
        fname = self.target[file_index]
        res = []
        with h5py.File(os.path.join(self.cwd, fname), 'r') as f:
            for key in f.keys():
                if 'xpcs' in key:
                    res.append(key)
        return res

    def get_twotime_qindex(self, ix, iy, hdl):
        res = twotime.get_twotime_qindex(self.meta, ix, iy, hdl)
        return res

    def plot_twotime_map(self, hdl, fname=None, **kwargs):
        if fname is None:
            fname = self.target[0]

        xfile = self.cache[fname]
        twotime.plot_twotime_map(xfile, hdl, meta=self.meta, **kwargs)
        return

    def plot_twotime(self, hdl, current_file_index=0, plot_index=1, **kwargs):
        if self.type != 'Twotime':
            self.show_message('Analysis type must be twotime.')
            return None

        fname = self.target[current_file_index]
        xfile = self.cache[fname]
        ret = twotime.plot_twotime(xfile, hdl, plot_index=plot_index,
                                   meta=self.meta, **kwargs)
        return ret

    def plot_intt(self, pg_hdl, max_points=128, rows=None, **kwargs):
        xf_list = self.get_xf_list(max_points, rows=rows)
        intt.plot(xf_list, pg_hdl, self.id_list, **kwargs)

    def plot_stability(self, mp_hdl, plot_id, **kwargs):
        fc = self.cache[self.target[plot_id]]
        stability.plot(fc, mp_hdl, **kwargs)

    def submit_job(self, *args, **kwargs):
        if len(self.target) <= 0:
            logger.error('no average target is selected')
            return
        worker = AverageToolbox(work_dir=self.cwd,
                                flist=self.target,
                                jid=self.avg_jid)
        worker.setup(*args, **kwargs)
        self.avg_worker.append(worker)
        logger.info('create average job, ID = %s', worker.jid)
        self.avg_jid += 1

        self.target.clear()
        return

    def remove_job(self, index):
        self.avg_worker.pop(index)
        return

    # def register_avg_worker(self, worker):
    #     g2_hist = np.zeros(worker.size, dtype=np.float32)
    #     self.avg_wo

    def update_avg_info(self, jid):
        self.avg_worker.layoutChanged.emit()
        if 0 <= jid < len(self.avg_worker):
            self.avg_worker[jid].update_plot()

    def update_avg_values(self, data):
        key, val = data[0], data[1]
        if self.avg_worker_active[key] is None:
            self.avg_worker_active[key] = [0, np.zeros(128, dtype=np.float32)]
        record = self.avg_worker_active[key]
        if record[0] == record[1].size:
            new_g2 = np.zeros(record[1].size * 2, dtype=np.float32)
            new_g2[0:record[0]] = record[1]
            record[1] = new_g2
        record[1][record[0]] = val
        record[0] += 1

        return
    
    def export_g2(self):
       pass 


if __name__ == "__main__":
    flist = os.listdir('./data')
    dv = ViewerKernel('./data', flist)



from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.backends.backend_qt5agg import NavigationToolbar2QT
from matplotlib.figure import Figure
from PyQt5 import QtGui, QtWidgets, QtCore
import numpy as np
import matplotlib.pyplot as plt
# matplotlib.pyplot.style.use(['science', 'no-latex'])

markers = ["o", "v", "^", "<", ">", "s", "p", "P", "*", "h", "H"]


class NavigationToolbarSimple(NavigationToolbar2QT):
    def __init__(self, *kw, **kwargs):
        super(NavigationToolbarSimple, self).__init__(*kw, **kwargs)

    def mouse_move(self, event):
        # just disable the mose_move event
        pass


# # # # # # class MplCanvasBarH(QtGui.QWidget):
class MplCanvasBarH(QtWidgets.QWidget):
    """
    MplWidget combines a MplCanvas with a vertical toolbar
    """
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbarSimple(self.hdl, self)
        self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.hbl = QtGui.QHBoxLayout()
        self.hbl.addWidget(self.hdl)
        self.hbl.addWidget(self.navi_toolbar)
        # self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.setLayout(self.hbl)
    
    def clear(self):
        self.hdl.clear()
        self.hdl.draw()


#  class MplCanvasBarV(QtGui.QWidget):
class MplCanvasBarV(QtWidgets.QWidget):
    """
    MplWidget combines a MplCanvas with a horizontal toolbar
    """
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbar2QT(self.hdl, self)
        self.vbl = QtGui.QVBoxLayout()
        self.vbl.addWidget(self.navi_toolbar)
        self.vbl.addWidget(self.hdl)
        self.setLayout(self.vbl)

    def clear(self):
        self.hdl.clear()
        self.hdl.draw()


# class MplCanvasBar(QtGui.QWidget):
class MplCanvasBar(QtWidgets.QWidget):
    """
    MplWidget combines a MplCanvas with a Toolbar
    """
    def __init__(self, parent=None):
        QtGui.QWidget.__init__(self, parent)
        self.hdl = MplCanvas()
        self.navi_toolbar = NavigationToolbarSimple(self.hdl, self)
        self.vbl = QtGui.QVBoxLayout()
        self.vbl.addWidget(self.hdl)
        self.vbl.addWidget(self.navi_toolbar)
        # self.navi_toolbar.setOrientation(QtCore.Qt.Vertical)
        self.setLayout(self.vbl)


class MplCanvas(FigureCanvasQTAgg):
    def __init__(self, parent=None, width=15, height=12, dpi=100):
        self.fig = Figure(figsize=(width, height), dpi=dpi)
        # self.axes = fig.add_subplot(111)
        super(MplCanvas, self).__init__(self.fig)
        self.shape = None
        self.axes = None
        self.obj = None

    def subplots(self, n, m, **kwargs):
        self.axes = self.fig.subplots(n, m, **kwargs)
        self.shape = (n, m)
        return self.axes

    def clear(self):
        self.clear_axes()
        self.fig.clear()
        self.axes = None
        self.obj = None
    
    def adjust_canvas_size(self, num_col, num_row):
        t = self.parent().parent().parent()
        if t is None:
            aspect = 1 / 1.618
        else:
            aspect = t.height() / self.width()

        min_size = t.height() - 20
        width = self.width()
        canvas_size = max(min_size, int(width / num_col * aspect * num_row))
        self.setMinimumSize(QtCore.QSize(0, canvas_size))

    def clear_axes(self):
        if self.axes is None:
            return
        else:
            if self.shape == (1, 1):
                self.axes.clear()
            else:
                for ax in self.axes.ravel():
                    ax.clear()

    def auto_scale(self, ylim=None, xlim=None, xscale=None, yscale=None):
        if self.axes is None:
            return
        else:
            for ax in np.array(self.axes).ravel():
                if xscale is not None:
                    ax.set_xscale(xscale)
                if yscale is not None:
                    ax.set_yscale(yscale)
                ax.relim()
                ax.autoscale_view(True, True, True)
                if ylim is not None:
                    ax.set_ylim(ylim)
                if xlim is not None:
                    ax.set_xlim(xlim)

    def update_lin(self, loc, x, y, visible=True):
        if self.obj is None:
            return
        lin_obj, = self.obj['lin'][loc]
        lin_obj.set_data(x, y)
        lin_obj.set_visible(visible)
        return

    def update_err(self, loc, x, y, y_error):
        if self.obj is None:
            return
        err_obj = self.obj['err'][loc]
        adjust_yerr(err_obj, x, y, y_error)
        return

    def show_image(self,
                   data,
                   vmin=None,
                   vmax=None,
                   extent=None,
                   cmap='seismic',
                   xlabel=None,
                   ylabel=None,
                   title=None,
                   id_list=None,
                   vline_freq=-1):
        def add_vline(ax, stop, vline_freq):
            if vline_freq < 0:
                return
            for x in np.arange(vline_freq, stop - 1, vline_freq):
                # for x in np.arange(1, stop // vline_freq - 1):
                ax.axvline(x - 0.5, ls='--', lw=0.5, color='black', alpha=0.5)

        if self.axes is None:
            ax = self.subplots(1, 1)
            add_vline(ax, data.shape[1], vline_freq)
            im0 = ax.imshow(data,
                            aspect='auto',
                            cmap=plt.get_cmap(cmap),
                            vmin=vmin,
                            vmax=vmax,
                            extent=extent,
                            interpolation=None)
            self.fig.colorbar(im0, ax=ax)

            ax.set_title(title)
            ax.set_xlabel(xlabel)
            ax.set_ylabel(ylabel)

            # when there are too many points, avoid labeling.
            # if data.shape < 20:
            #     ax.set_xticks(np.arange(data.shape[1]))
            #     ax.set_xticklabels(id_list[0: data.shape[1]])

            self.obj = [im0]
            self.fig.tight_layout()
        else:
            self.obj[0].set_data(data)
            self.obj[0].set_clim(vmin, vmax)
            self.axes.set_title(title)
            self.axes.set_xlabel(xlabel)
            self.axes.set_ylabel(ylabel)

        self.draw()
        return

    def show_lines(self,
                   data,
                   xlabel=None,
                   ylabel=None,
                   title=None,
                   legend=None,
                   loc='best',
                   rows=None
                   ):

        if legend in [None, False]:
            legend = np.arange(len(data))

        if rows in [None, []]:
            alpha = np.ones(len(data)) * 0.75
        else:
            alpha = np.ones(len(data)) * 0.15
            for t in rows:
                if t < len(data):
                    alpha[t] = 1.0

        if isinstance(data, np.ndarray):
            x = np.arange(data.shape[1])
            data2 = []
            for n in range(data.shape[0]):
                data2.append([x, data[n]])
            data = data2

        if self.axes is None or len(data) != len(self.obj):
            ax = self.subplots(1, 1)
            line_obj = []
            for n in range(len(data)):
                mk = markers[n % len(markers)]
                line = ax.plot(data[n][0], data[n][1], mk + '--', ms=3,
                               alpha=alpha[n], label=legend[n])
                line_obj.append(line)
            self.obj = line_obj

            if legend is not None:
                ax.legend(loc=loc)

        else:
            for n in range(len(data)):
                line, = self.obj[n]
                line.set_data(data[n][0], data[n][1])
                if legend is not None:
                    line.set_label(legend[n])
            self.auto_scale()

        self.axes.set_title(title)
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel(ylabel)
        self.draw()
        return

    def show_scatter(self,
                     data,
                     color=None,
                     xlabel=None,
                     ylabel=None,
                     title=None,
                     legend=None,
                     loc='best',
                     alpha=0.85):
        if data.ndim != 2 or data.shape[0] != 2:
            raise ValueError('input data shape not supported')
        x, y = data[0], data[1]
        if color is None:
            color = np.arange(x.size)
        # if legend in [None, False]:
        #     legend = np.arange(len(x))

        if self.axes is not None:
            self.clear()
        ax = self.subplots(1, 1)
        line = ax.scatter(x, y, c=color)
        self.fig.colorbar(line, ax=ax)
        if legend is not None:
            ax.legend(loc=loc)

        self.axes.set_title(title)
        self.axes.set_xlabel(xlabel)
        self.axes.set_ylabel(ylabel)
        self.draw()
        return


# https://github.com/matplotlib/matplotlib/issues/4556
def adjust_yerr(err_obj, x, y, y_error):
    # not using error top / bot bar;
    # ln, (err_top, err_bot), (bars, ) = err_obj
    ln, _, (bars, ) = err_obj
    ln.set_data(x, y)

    yerr_top = y + y_error
    yerr_bot = y - y_error

    # err_top.set_ydata(yerr_top)
    # err_bot.set_ydata(yerr_bot)

    new_segments = [
        np.array([[x, yt], [x, yb]])
        for x, yt, yb in zip(x, yerr_top, yerr_bot)
    ]

    bars.set_segments(new_segments)


MplToolbar = NavigationToolbar2QT
# class MplToolbar(MplCanvas):
#     def __init__(self, **kwargs):
#         super(MplCanvas, self).__init__(**kwargs)
#         toolbar = NavigationToolbar2QT(self.fig, self)
#         layout = QtWidgets.QVBoxLayout()
#         layout.addWidget(toolbar)
#         layout.addWidget(self)
#
#         widget = QtWidgets.QWidget()
#         widget.setLayout(layout)
#         self.setCentralWidget(widget)


key = {
  "saxs_1d": "/exchange/partition-mean-total",
  "Iqp": "/exchange/partition-mean-partial",
  "ql_sta": "/xpcs/sqlist",
  "ql_dyn": "/xpcs/dqlist",
  "dqmap": "/xpcs/dqmap",
  "sqmap": "/xpcs/sqmap",
  "type": "/xpcs/analysis_type",
  "t0": "/measurement/instrument/detector/exposure_period",
  "t1": "/measurement/instrument/detector/exposure_time",
  "tau": "/exchange/tau",
  "g2": "/exchange/norm-0-g2",
  "g2_err": "/exchange/norm-0-stderr",
  "saxs_2d": "/exchange/pixelSum",
  "Int_t": "/exchange/frameSum",
  "ccd_x0": "/measurement/instrument/acquisition/beam_center_x",
  "ccd_y0": "/measurement/instrument/acquisition/beam_center_y",
  "det_dist": "/measurement/instrument/detector/distance",
  "pix_dim": "/measurement/instrument/detector/x_pixel_size",
  "X_energy": "/measurement/instrument/source_begin/energy",
  "xdim": "/measurement/instrument/detector/x_dimension",
  "ydim": "/measurement/instrument/detector/y_dimension",
  "c2_half": "/exchange/C2T_all",
  "g2_full": "/exchange/g2full",
  "g2_partials": "/exchange/g2partials",
  "G2": "/exchange/G2",
  "IP": "/exchange/IF",
  "IF": "/exchange/IP"
}



import h5py
import numpy as np
from scipy.stats import describe
import os
np.set_printoptions(precision=3, suppress=True)


def c2r(x):
    # def convert_to_readable_number(x):
    if not isinstance(x, np.ndarray):
        x = np.array(x)

    val = np.max(np.abs(x))
    if val > 1E4 or val < 1E-2:
        if x.size == 1:
            res = '%.3e' % x
        else:
            res_list = ['%.3e' % t for t in x]
            res = '[' + ', '.join(res_list) + ']'
    else:
        res = str(np.round(x, 3))
    return res


def describe_numpy(arr):
    repr = str(arr.shape) + ', ' + str(arr.dtype) + ':'
    if arr.size > 1:
        res = describe(arr.ravel())
        repr += 'minmax = %s, mean = %s' % (c2r(res.minmax),
                                            c2r(res.mean))
    else:
        repr += 'val = %s' % c2r(arr[0])
    return repr


def read_h5py(hdl, path, level, guide_str0):
    if path not in hdl:
        return None
    if 'C2T_all' in path:
        return ['C2T_all']
    result = []

    guide_mid = guide_str0 + '├──'
    guide_lst = guide_str0 + '└──'
    guide_nxt_mid = guide_str0 + '│  '
    guide_nxt_lst = guide_str0 + '   '

    if isinstance(hdl[path], h5py._hl.dataset.Dataset):

        value = hdl[path][()]
        # there are 3 kinds of strings in the hdf file;
        if isinstance(value, (np.bytes_, bytes, str)):
            if not isinstance(value, str):
                value = hdl[path][()].decode() 
            return [guide_lst + value]

        if hdl[path].shape == ():
            return [guide_lst + 'empty']

        info = describe_numpy(np.array(hdl[path]))
        result.append(guide_lst + info)
        return result

    for n, key in enumerate(hdl[path].keys()):
        if n == len(hdl[path].keys()) - 1:
            guide = guide_lst
            guide_nxt = guide_nxt_lst
        else:
            guide = guide_mid
            guide_nxt = guide_nxt_mid

        # os.path.join won't work on windows systems
        # new_path = os.path.join(path, key)
        new_path = '/'.join([path, key])
        if isinstance(key, str):
            result.append(guide + key)
            info = read_h5py(hdl, new_path, level + 1, guide_nxt)
            if info is not None:
                result = result + info
        elif isinstance(key, np.ndarray):
            info = describe_numpy(hdl[new_path])
            result.append(guide + info)
        else:
            pass

    return result


def get_hdf_info(path, fname):
    with h5py.File(os.path.join(path, fname), 'r') as hdl:
        res = read_h5py(hdl, '.', 0, '')
    return res


if __name__ == '__main__':
    pass




import numpy as np
import pyqtgraph as pg

colors = [
    (192, 0, 0),
    (0, 176, 80),
    (0, 32, 96),
    (255, 0, 0),
    (0, 176, 240),
    (0, 32, 96),
    (255, 164, 0),
    (146, 208, 80),
    (0, 112, 192),
    (112, 48, 160),
    (54, 96, 146),
    (150, 54, 52),
    (118, 147, 60),
    (96, 73, 122),
    (49, 134, 155),
    (226, 107, 10),
]


def smooth_data(fc, window=1, sampling=1):
    # some bad frames have both x and y = 0;
    # x, y = fc.Int_t[0], fc.Int_t[1]
    y = fc.Int_t[1]
    x = np.arange(y.shape[0])

    if window > 1:
        y = np.cumsum(y, dtype=float, axis=0)
        y = (y[window:] - y[:-window]) / window
        x = x[window:]
    if sampling >= 2:
        y = y[::sampling]
        x = x[::sampling]

    return x, y


def matplot_plot(xf_list, pg_hdl, legend, rows=None, **kwargs):
    data = []
    for fc in xf_list:
        x, y = smooth_data(fc, **kwargs)
        data.append([x, y])

    pg_hdl.clear()
    pg_hdl.show_lines(data,
                      xlabel="Frame Index",
                      ylabel="Intensity (ph/pixel)",
                      loc='lower right',
                      legend=legend,
                      rows=rows)
    pg_hdl.draw()


def plot(xf_list, pg_hdl, enable_zoom=True, xlabel='Frame Index', **kwargs):
    """
    :param xf_list: list of xf objects
    :param pg_hdl: pyqtgraph handler to plot
    :param enable_zoom: bool, if to plot the zoom view or not
    :param xlabel:
    :param kwargs: used to define how to average/sample the data
    :return:
    """

    t0 = xf_list[0].t0
    data = []
    for fc in xf_list:
        x, y = smooth_data(fc, **kwargs)
        if xlabel != 'Frame Index':
            x = x * t0
        data.append([x, y])

    pg_hdl.clear()

    t = pg_hdl.addPlot(colspan=2)
    t.addLegend(offset=(-1, 1), labelTextSize='8pt', verSpacing=-10)
    tf = pg_hdl.addPlot(row=1, col=0, title='Fourier Spectrum')
    tf.addLegend(offset=(-1, 1), labelTextSize='8pt', verSpacing=-10)
    tf.setLabel('bottom', 'Frequency (Hz)')
    tf.setLabel('left', 'FFT Intensity')

    tz = pg_hdl.addPlot(row=1, col=1, title='Zoom In')
    tz.addLegend(offset=(-1, 1), labelTextSize='8pt', verSpacing=-10)

    t.setDownsampling(mode='peak')
    tf.setDownsampling(mode='peak')
    tz.setDownsampling(mode='peak')

    for n in range(len(data)):
        t.plot(data[n][0], data[n][1], pen=pg.mkPen(colors[n], width=1),
               name=xf_list[n].label)
    t.setTitle('Intensity vs %s' % xlabel)

    if enable_zoom:
        vmin = np.min(data[0][0])
        vmax = np.max(data[0][0])
        cen = vmin * 0.382 + vmax * 0.618
        width = (vmax - vmin) * 0.05
        lr = pg.LinearRegionItem([cen - width, cen + width])
        # lr.setZValue(-10)
        t.addItem(lr)
    t.setLabel('bottom', '%s' % xlabel)
    t.setLabel('left', 'Intensity (cts / pixel)')

    for n in range(len(data)):
        y = np.abs(np.fft.fft(data[n][1]))

        x = np.arange(y.size // 2) / (y.size * t0 * kwargs['sampling'])
        y = y[slice(0, y.size // 2)]
        # get ride of zero frequency;
        y[0] = 0

        tf.plot(x, y, pen=pg.mkPen(colors[n], width=1), name=xf_list[n].label)

    for n in range(len(data)):
        tz.plot(data[n][0], data[n][1], pen=pg.mkPen(colors[n], width=1),
                name=xf_list[n].label)

    def update_plot():
        tz.setXRange(*lr.getRegion(), padding=0)

    def update_region():
        lr.setRegion(tz.getViewBox().viewRange()[0])

    lr.sigRegionChanged.connect(update_plot)
    tz.sigXRangeChanged.connect(update_region)

    tz.setLabel('bottom', '%s' % xlabel)
    tz.setLabel('left', 'Intensity (cts / pixel)')
    update_plot()

    return



import numpy as np
import scipy as sp
from scipy.optimize import curve_fit
from scipy.stats import linregress as sp_stats
import traceback
from sklearn import linear_model
import joblib
import os
import time
import traceback
import logging


logger = logging.getLogger(__name__)

cache_dir = os.path.join(os.path.expanduser('~'), '.xpcs_viewer')
from joblib import Memory
memory = Memory(cache_dir, verbose=0)


@memory.cache
def fit_with_fixed(*args, **kwargs):
    # wrap the fitting function in memory so avoid re-run
    return fit_with_fixed_raw(*args, **kwargs)


def single_exp(x, tau, bkg, cts):
    return cts * np.exp( -2 * x / tau) + bkg


def fit_tau(qd, tau, tau_err):
    x = np.log(qd).reshape(-1, 1)
    y = np.log(tau).reshape(-1, 1)
    dy = tau / tau_err
    reg = linear_model.LinearRegression()
    reg.fit(x, y, sample_weight=dy)
    x2 = np.linspace(np.min(x) - 0.1, np.max(x) + 0.1, 128)
    y2 = reg.predict(x2.reshape(-1, 1))
    return reg.coef_, reg.intercept_, np.exp(x2).ravel(), np.exp(y2).ravel()


def fit_xpcs(tel, qd, g2, g2_err, b):
    """
    :param tel: t_el
    :param qd: ql_dyn
    :param g2: g2 [time, ql_dyn]
    :param g2_err: [time, ql_dyn]
    :param b: bounds
    :return:
    """

    # fit_x = np.logspace(-5, 0.5, num=128)
    fit_x = np.logspace(np.log10(np.min(tel)) - 0.5,
                        np.log10(np.max(tel)) + 0.5, 128)

    p0_guess = [np.sqrt(b[0][0] * b[1][0]),
                0.5 * (b[0][1] + b[1][1]),
                0.5 * (b[0][2] + b[1][2])]

    fit_val = np.zeros(shape=(qd.size, 7))
    fit_result = []
    for n in range(qd.size):
        err = g2_err[:, n]
        result = {'num_zero_err': np.sum(err < 1E-6)}
        avg = np.mean(err[err > 1E-6])
        err[err <= 1E-6] = avg
        fit_val[n, 0] = qd[n]

        try:
            popt, pcov = curve_fit(single_exp, tel, g2[:, n],
                                   p0=p0_guess,
                                   sigma=err,
                                   bounds=b)
            fit_val[n, 1:4], fit_val[n, 4:7] = popt, np.sqrt(np.diag(pcov))
        except:
            # fit_val[n, 1:4], fit_val[n, 4:7] = popt, np.sqrt(np.diag(pcov))
            result = {'err_msg': 'q_index %2d:' + str(traceback.format_exc()),
                      'fit_x': fit_x, 'fit_y': np.ones_like(fit_x)}
        else:
            fit_y = single_exp(fit_x, *popt)
            result = {'err_msg': None,
            # result = {'err_msg': 'q_index %2d: fit ends without err' % n,
                      'opt': popt, 'err': np.sqrt(np.diag(pcov)),
                      'fit_x': fit_x, 'fit_y': fit_y}
        finally:
            fit_result.append(result)

    return fit_result, fit_val


def fit_with_fixed_raw(base_func, x, y, sigma, bounds, fit_flag, fit_x,
                       p0=None):
    """
    :param base_func: the base function used for fitting; it can have multiple
        input variables, some of which can be fixed during the fitting;
    :param x: scaler input
    :param y: scaler output
    :param sigma: the error for y value
    :param bounds: tuple with two elements. 1st is the lower bounds and 2nd is
        the upper bounds; if the fit_flag for a variable is False, then the
        upper bound is used as the fixed value;
    :param fit_flag: tuple of bools, True/False for fit and fixed
    :param fit_x: the fitting line for x
    :param p0: the initial value for the variables; if None is provided, the
        intial value is set as the mean of lower and upper bounds
    :return: a tuple of (fit_line, fit_val)
    """
    if not isinstance(fit_flag, np.ndarray):
        fit_flag = np.array(fit_flag)

    fix_flag = np.logical_not(fit_flag)

    if not isinstance(bounds, np.ndarray):
        bounds = np.array(bounds)

    # degree of fitting
    # dof = np.sum(fit_flag)

    # number of arguments, regardless of fixed or to be fitted
    num_args = len(fit_flag)

    # create a function that takes care of the fit flag;
    def func(x1, *args):
        inputs = np.zeros(num_args)
        inputs[fix_flag] = bounds[1, fix_flag]
        inputs[fit_flag] = np.array(args)
        return base_func(x1, *inputs)

    # process boundaries and initial values
    bounds_fit = bounds[:, fit_flag]
    # doing a simple average to get the initial guess;
    if p0 is None:
        p0 = np.mean(bounds_fit, axis=0)
    else:
        p0 = np.array(p0)[fit_flag]

    fit_val = np.zeros((y.shape[1], 2, num_args))

    fit_line = []
    for n in range(y.shape[1]):
        flag = True
        try:
            popt, pcov = curve_fit(func, x, y[:, n], p0=p0, sigma=sigma[:, n],
                                   bounds=bounds_fit)
        except (Exception, RuntimeError, ValueError, Warning) as err:
            msg = "Fitting failed: %s" % traceback.format_exc()
            logger.info(msg)
            flag = False
            fit_val[n, 0, fit_flag] = p0 
            fit_val[n, 0, fix_flag] = bounds[1, fix_flag]
            # mark failed fitting to be negative so they can be filtered later
            fit_val[n, 1, :] = -1
            fit_y = None

        else:
            flag = True
            msg = 'FittingSuccess'
            # converge values
            fit_val[n, 0, fit_flag] = popt
            fit_val[n, 0, fix_flag] = bounds[1, fix_flag]
            # errors; the fixed variables have error of 0
            fit_val[n, 1, fit_flag] = np.sqrt(np.diag(pcov))
            # fit line
            fit_y = func(fit_x, *popt)

        finally:
            fit_line.append({'fit_x': fit_x, 'fit_y': fit_y, 'success': flag,
                             'msg': msg})

    return fit_line, fit_val



import pyqtgraph as pg
from pyqtgraph import ImageView, PlotWidget, GraphicsLayoutWidget
from .mpl_cmaps_in_ImageItem import pg_get_cmap
import matplotlib.pyplot as plt
from PyQt5 import QtCore


pg.setConfigOptions(imageAxisOrder='row-major')


class ImageViewDev(ImageView):
    def __init__(self, *args, **kwargs) -> None:
        super(ImageViewDev, self).__init__(*args, **kwargs)

    def adjust_viewbox(self):
        vb = self.getView()
        xMin, xMax = vb.viewRange()[0]
        yMin, yMax = vb.viewRange()[1]

        vb.setLimits(xMin=xMin,
                     xMax=xMax,
                     yMin=yMin,
                     yMax=yMax,
                     minXRange=(xMax - xMin) / 50,
                     minYRange=(yMax - yMin) / 50)
        vb.setMouseMode(vb.RectMode)
        vb.setAspectLocked(1.0)

    def reset_limits(self):
        """
        reset the viewbox's limits so updating image won't break the layout;
        """
        self.view.state['limits'] = {'xLimits': [None, None],
                                     'yLimits': [None, None],
                                     'xRange': [None, None],
                                     'yRange': [None, None]
                                     }

    def set_colormap(self, cmap):
        pg_cmap = pg_get_cmap(plt.get_cmap(cmap))
        self.setColorMap(pg_cmap)

    def add_readback(self, display=None, extent=None, type='log'):
        # vLine = pg.InfiniteLine(angle=90, movable=False)
        # hLine = pg.InfiniteLine(angle=0, movable=False)
        # self.view.addItem(vLine, ignoreBounds=True)
        # self.view.addItem(hLine, ignoreBounds=True)

        # def print_roi_shape(evt):
        #     print(self.roi.boundingRect())

        # self.roi.sigRegionChanged.connect(print_roi_shape)

        def compute_qxy(col, row):
            s = self.image.shape[-2:]
            # qx
            a1, a2 = col, s[1] - col
            qx = (extent[0] * a2 + extent[1] * a1) / s[1]

            # qy
            b1, b2 = row, s[0] - row
            qy = (extent[2] * b2 + extent[3] * b1) / s[0]

            return qx, qy

        def mouse_moved(pos):
            shape = self.image.shape
            if self.scene.itemsBoundingRect().contains(pos):
                mouse_point = self.getView().mapSceneToView(pos)
                # vLine.setPos(mouse_point.x())
                # hLine.setPos(mouse_point.y())
                col = int(mouse_point.x())
                row = int(mouse_point.y())

                if col < 0 or col >= shape[-1]:
                    return
                if row < 0 or row >= shape[-2]:
                    return

                if len(shape) == 3:
                    pixel_val = self.image[self.currentIndex, row, col]
                elif len(shape) == 2:
                    pixel_val = self.image[row, col]
                else:
                    raise ValueError('Check array dimension')

                if type == 'log':
                    pixel_val = 10 ** pixel_val
                qx, qy = compute_qxy(col, row)

                if display is None:
                    print(pixel_val)
                else:
                    display.clear()
                    display.setText(
                        '%d: [x=%4d, y=%4d, qx=%fÅ⁻¹, qy=%fÅ⁻¹, c:%.3f]' % (
                            self.currentIndex, col, row, qx, qy, pixel_val))
        self.scene.sigMouseMoved.connect(mouse_moved)

    def clear(self):
        super(ImageViewDev, self).clear()
        self.reset_limits()
        # incase the signal isn't connected to anything.
        try:
            self.scene.sigMouseMoved.disconnect()
        except:
            pass


class PlotWidgetDev(GraphicsLayoutWidget):
    def __init__(self, *args, **kwargs) -> None:
        super(PlotWidgetDev, self).__init__(*args, **kwargs)
        self.setBackground('w')

    def adjust_canvas_size(self, num_col, num_row):
        t = self.parent().parent().parent()
        if t is None:
            aspect = 1 / 1.618
        else:
            aspect = t.height() / self.width()

        min_size = t.height() - 20
        width = self.width()
        canvas_size = max(min_size, int(width / num_col * aspect * num_row))
        self.setMinimumSize(QtCore.QSize(0, canvas_size))



import sys


# https://stackoverflow.com/questions/19425736
# /how-to-redirect-stdout-and-stderr-to-logger-in-python
class LoggerWriter:
    def __init__(self, level):
        # self.level is really like using log.debug(message)
        # at least in my case
        self.level = level

    def write(self, message):
        # if statement reduces the amount of newlines that are
        # printed to the logger
        if message != '\n':
            self.level(message)

    def flush(self):
        # create a flush method so things can be flushed when
        # the system wants to. Not sure if simply 'printing'
        # sys.stderr is the correct way to do it, but it seemed
        # to work properly for me.
        self.level(sys.stderr)
